pragma solidity ^0.4.22;

import "strings.sol";

library ChainDataCmdHelper {
 
    enum Base {BASE16, BASE58, BASE64}
    enum Hash {SHA2_256, SHA3_256, KECCAK256}
    enum Content {MYCHAIN010_BLOCK, MYCHAIN010_TX, MYCHAIN010_HEADER}
 
    bytes1 constant udag_cid_version = hex'01';
 
    string constant base16_codec = "f";
    string constant base58_codec = "b";
    string constant base64_codec = "m";
 
    bytes1 constant sha2_256_codec = hex'32';
    bytes1 constant sha3_256_codec = hex'35';
    bytes1 constant keccak256_codec = hex'39';  
 
    bytes1 constant MYCHAIN010_block_codec = hex'a5';
    bytes1 constant MYCHAIN010_tx_codec = hex'a6';
    bytes1 constant MYCHAIN010_header_codec = hex'a8';    
 
    /**
     * @dev build mychain010 block content id
     * @param _hash bytes
     */  
    function buildMychain010BlockCID(bytes _hash) internal pure returns (string){    
        uint8 hash_length = uint8(_hash.length);
        bytes memory multihash = concat(concat(varintEncoding(uint8(sha2_256_codec)), varintEncoding(hash_length)), _hash);
        bytes memory multicontent = concat(concat(varintEncoding(uint8(udag_cid_version)), varintEncoding(uint8(MYCHAIN010_block_codec))), multihash);
        string memory multibase = bytesToHexString(multicontent);
        string memory content_id = strJoin(base16_codec, multibase);
        return content_id;        
    }
 
    /**
     * @dev build mychain010 tx content id
     * @param _hash bytes
     */  
    function buildMychain010TxCID(bytes _hash) internal pure returns (string){
        uint8 hash_length = uint8(_hash.length);
        bytes memory multihash = concat(concat(varintEncoding(uint8(sha2_256_codec)), varintEncoding(hash_length)), _hash);
        bytes memory multicontent = concat(concat(varintEncoding(uint8(udag_cid_version)), varintEncoding(uint8(MYCHAIN010_tx_codec))), multihash);
        string memory multibase = bytesToHexString(multicontent);
        string memory content_id = strJoin(base16_codec, multibase);
        return content_id;        
    }
 
    /**
     * @dev build mychain010 blockheader content id
     * @param _hash bytes
     */  
    function buildMychain010HeaderCID(bytes _hash) internal pure returns (string memory){
        uint8 hash_length = uint8(_hash.length);
        bytes memory multihash = concat(concat(varintEncoding(uint8(sha2_256_codec)), varintEncoding(hash_length)), _hash);
        bytes memory multicontent = concat(concat(varintEncoding(uint8(udag_cid_version)), varintEncoding(uint8(MYCHAIN010_header_codec))), multihash);
        string memory multibase = bytesToHexString(multicontent);
        string memory content_id = strJoin(base16_codec, multibase);
        return content_id;        
    }
 
    /**
     * @dev build udag command
     * @param _domain string
     * @param _cid string
     * @param _path string
     * @param _parser_cmd string
     * 
     */
    function buildCommand(string _domain, string _cid, string _path, string _parser_cmd) internal pure returns(string){
 
        string memory udag_cmd;
        string memory udag_url;
        string memory udag_url_path;
 
        var path_slice = strings.toSlice(_path);
        var parser_slice = strings.toSlice(_parser_cmd);
        var slash_slice = strings.toSlice("/");
 
        if(strings.empty(path_slice) && strings.empty(parser_slice)){
            udag_cmd = strJoin("udag://", _domain, "/", _cid);
        }
        else if(strings.empty(path_slice)){
            udag_url = strJoin("udag://", _domain, "/", _cid);
            udag_cmd = strJoin(udag_url," --json-path ", _parser_cmd);
        }
        else if(strings.empty(parser_slice)){
            udag_url = strJoin("udag://", _domain, "/", _cid);
 
            if(strings.startsWith(path_slice, slash_slice)){
                udag_cmd = strJoin(udag_url, _path);
            }else{
                udag_cmd = strJoin(udag_url,"/", _path);
            }
 
        }
        else{
            udag_url = strJoin("udag://", _domain, "/", _cid);
            if(strings.startsWith(path_slice, slash_slice)){
                udag_url_path = strJoin(udag_url, _path);
            }else{
                udag_url_path = strJoin(udag_url,"/", _path);
            }
            udag_cmd = strJoin(udag_url_path," --json-path ", _parser_cmd);
        }
 
        return udag_cmd;
    }
 
    /**
     * @dev generate udag content id
     * @param _hash bytes
     * @param _hash_codec UDAGHashList
     * @param _content_codec UDAGContentList
     * @param _base_codec UDAGBaseList
     */  
    function generateCID(bytes _hash,
                        Hash _hash_codec,
                        Content _content_codec,
                        Base _base_codec) internal pure returns (string){
 
        string memory content_id;
        uint8 hash_length = uint8(_hash.length);
        bytes1 hash_codec;
        bytes1 content_codec;
        string memory base_codec;
 
        if (_hash_codec == Hash.SHA2_256){
            hash_codec = sha2_256_codec;
        }
        else if (_hash_codec == Hash.SHA3_256){
            hash_codec = sha3_256_codec;
        }
        else if(_hash_codec == Hash.KECCAK256){
            hash_codec = keccak256_codec;
        }
 
        if (_base_codec == Base.BASE16){
            base_codec = base16_codec;
        }
        else if (_base_codec == Base.BASE58){
            base_codec = base58_codec;
        }
        else if(_base_codec == Base.BASE64){
            base_codec = base64_codec;
        }
 
        if (_content_codec == Content.MYCHAIN010_BLOCK){
            content_codec = MYCHAIN010_block_codec;
        }
        else if(_content_codec == Content.MYCHAIN010_TX){
            content_codec = MYCHAIN010_tx_codec;
        }
        else if (_content_codec == Content.MYCHAIN010_HEADER){
            content_codec = MYCHAIN010_header_codec;
        }
 
        bytes memory multihash = concat(concat(varintEncoding(uint8(hash_codec)), varintEncoding(hash_length)), _hash);
        bytes memory multicontent = concat(concat(varintEncoding(uint8(udag_cid_version)), varintEncoding(uint8(content_codec))), multihash);
        string memory multibase;
 
        if(_base_codec == Base.BASE16){
            multibase = bytesToHexString(multicontent);
 
        }
        else if(_base_codec == Base.BASE64){
            bool ret_code;
            (ret_code, multibase) = base64_encode(bytesToHexString(multicontent));
 
            if(!ret_code){
                revert("BASE_ENCODE_ERROR: Registering oracle node failed. encoding base64 AVR failed");                    
            }
        }
 
        content_id = strJoin(base_codec, multibase);
        return content_id;
    }
 
    /**
     * @dev encode unsigned varint 
     * @param _n uint8
     */  
    function varintEncoding(uint8 _n) internal pure returns (bytes){
 
        bytes memory varint1;
        bytes memory varint2;
 
        for (uint32 i = 0; i < 2; i++) {
            uint8 a = (_n & 0x7F);
            _n = _n >> 7;
            if (_n == 0) {
                varint1 = uint8ToBytes(a);
 
                if(i == 0){
                    return varint1;
                }
                else{
                    return concat(varint2, varint1);
                }
 
            } else {
                varint2 = uint8ToBytes(a | 0x80);
            }
        }
    }
 
    function concat(bytes memory _preBytes, bytes memory _postBytes) internal pure returns (bytes) {
        bytes memory tempBytes;
 
        assembly {
 
            tempBytes := mload(0x40)
            let length := mload(_preBytes)
            mstore(tempBytes, length)
            let mc := add(tempBytes, 0x20)
            let end := add(mc, length)
 
            for {
 
                let cc := add(_preBytes, 0x20)
            } lt(mc, end) {
                mc := add(mc, 0x20)
                cc := add(cc, 0x20)
            } {
 
                mstore(mc, mload(cc))
            }
 
            length := mload(_postBytes)
            mstore(tempBytes, add(length, mload(tempBytes)))
 
            mc := end
            end := add(mc, length)
 
            for {
                let cc := add(_postBytes, 0x20)
            } lt(mc, end) {
                mc := add(mc, 0x20)
                cc := add(cc, 0x20)
            } {
                mstore(mc, mload(cc))
            }
 
            mstore(0x40, and(
              add(add(end, iszero(add(length, mload(_preBytes)))), 31),
              not(31) // Round down to the nearest 32 bytes.
            ))
        }
 
        return tempBytes;
    }
 
    function uint8ToBytes(uint8 input) internal pure returns (bytes) {
        bytes memory b = new bytes(1);
        byte temp = byte(input);
        b[0] = temp;
        return b;
    }
 
    function bytesToHexString(bytes bs) internal pure returns(string) {
        bytes memory tempBytes = new bytes(bs.length * 2);
        uint len = bs.length;
        for (uint i = 0; i < len; i++) {
            byte b = bs[i];
            byte nb = (b & 0xf0) >> 4;
            tempBytes[2 * i] = nb > 0x09 ? byte((uint8(nb) + 0x37)) : (nb | 0x30);
            nb = (b & 0x0f);
            tempBytes[2 * i + 1] = nb > 0x09 ? byte((uint8(nb) + 0x37)) : (nb | 0x30);
        }
        return string(tempBytes);
    }
 
    function strJoin(string _a, string _b, string _c, string _d, string _e) internal pure returns (string){
        bytes memory _ba = bytes(_a);
        bytes memory _bb = bytes(_b);
        bytes memory _bc = bytes(_c);
        bytes memory _bd = bytes(_d);
        bytes memory _be = bytes(_e);
        string memory abcde = new string(_ba.length + _bb.length + _bc.length + _bd.length + _be.length);
        bytes memory babcde = bytes(abcde);
        uint k = 0;
        for (uint i = 0; i < _ba.length; i++) babcde[k++] = _ba[i];
        for (i = 0; i < _bb.length; i++) babcde[k++] = _bb[i];
        for (i = 0; i < _bc.length; i++) babcde[k++] = _bc[i];
        for (i = 0; i < _bd.length; i++) babcde[k++] = _bd[i];
        for (i = 0; i < _be.length; i++) babcde[k++] = _be[i];
        return string(babcde);
    }
 
    function strJoin(string _a, string _b, string _c, string _d) internal pure returns (string) {
        return strJoin(_a, _b, _c, _d, "");
    }
 
    function strJoin(string _a, string _b, string _c) internal pure returns (string) {
        return strJoin(_a, _b, _c, "", "");
    }
 
    function strJoin(string _a, string _b) internal pure returns (string) {
        return strJoin(_a, _b, "", "", "");
    }
}